﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;
using CatanGame.ViewModel;
using System.ComponentModel;

namespace CatanGame.View
{
    public class CatanBoardCanvas : Canvas
    {
        #region Private fields
        private CatanViewModel _viewModel;
        #endregion

        #region Public constructors
        public CatanBoardCanvas() : base()
        { 
            this.Loaded += (obj, args) => {
                InitDraw();
                SizeChanged += (obj, args) => DrawMap();
            };
        }
        #endregion

        #region Private fields


        private Ellipse[,] _roads;
        private Ellipse[,,] _settlements;
        private CatanField[,] _fields;
        #endregion

        #region Public Properties
        public int Offset { get => _viewModel.BoardSize / 2; }
        public double EllipseDiameter { get; set; }
        public double WindowSize { get => Math.Min(ActualHeight, ActualWidth); }
        public double HexagonRadius { get => WindowSize / (_viewModel.BoardSize * 2); }

        // dependencyProperty functionality taken from wpf github: https://github.com/dotnet/wpf/blob/main/src/Microsoft.DotNet.Wpf/src/PresentationFramework/System/Windows/Controls/ItemsControl.
        public static readonly DependencyProperty CatanBoardTemplateProperty =
        DependencyProperty.Register(
                "CatanBoardTemplate",
                typeof(DataTemplate),
                typeof(ItemsControl));

        [Bindable(true)]
        public DataTemplate CatanBoardTemplate { get => (DataTemplate) GetValue(CatanBoardTemplateProperty); set => SetValue(CatanBoardTemplateProperty, value); }

        public static readonly DependencyProperty FieldsSourceProperty =
            DependencyProperty.Register(
                "FieldsSource",
                typeof(IEnumerable),
                typeof(ItemsControl));

        [Bindable(true)]
        public IEnumerable FieldsSource
        {
            get { return (IEnumerable) GetValue(FieldsSourceProperty); }
            set
            {
                if (value == null)
                {
                    ClearValue(FieldsSourceProperty);
                }
                else
                {
                    SetValue(FieldsSourceProperty, value);
                }
            }
        }
        public CatanViewModel ViewModel 
        { 
            get => _viewModel;
            set
            {
                _viewModel = value;
                _roads = new Ellipse[_viewModel.BoardSize * _viewModel.BoardSize * 2, _viewModel.BoardSize * _viewModel.BoardSize * 2];
                _settlements = new Ellipse[_viewModel.BoardSize, _viewModel.BoardSize, 2];
                _fields = new CatanField[_viewModel.BoardSize, _viewModel.BoardSize];
                InitDraw();
            }
        }

        #endregion

        #region Private methods
        private Brush GetBrush(String fieldType)
        {
            switch (fieldType)
            {
                case "None":
                    return Brushes.SandyBrown;
                case "Lumber":
                    return Brushes.DarkOliveGreen;
                case "Brick":
                    return Brushes.SaddleBrown;
                case "Wool":
                    return Brushes.YellowGreen;
                case "Grain":
                    return Brushes.Yellow;
                case "Ore":
                    return Brushes.Gray;
                case "Sea":
                    return Brushes.LightBlue;
                default:
                    throw new InvalidOperationException("Invalid resource type!");
            }
        }

        private void InitDraw()
        {
            if (_viewModel is null || HexagonRadius == 0) return;

            Children.Clear();

            for (int q = 0; q < _viewModel.BoardSize; ++q)
            {
                for (int r = 0; r < _viewModel.BoardSize; ++r)
                {
                    if (_viewModel.Fields[r * _viewModel.BoardSize + q] is not null)
                    {
                        Point position = CalculatePositionOnCanvas(r, q);

                        CreateFields(position, (r, q));
                        CreateRoadSpots(position, (r, q));
                    }
                }
            }
            CreateBuildingSpots();
        }

        private void DrawMap()
        {
            if (_viewModel is null || HexagonRadius == 0) return;

            for (int q = 0; q < _viewModel.BoardSize; ++q)
            {
                for (int r = 0; r < _viewModel.BoardSize; ++r)
                {
                    if (_viewModel.Fields[r * _viewModel.BoardSize + q] is not null)
                    {
                        Point position = CalculatePositionOnCanvas(r, q);

                        UpdateFields(position, (r, q));
                        UpdateRoadSpots(position, (r, q));
                    }
                }
            }
            UpdateBuildingSpots();
        }

        private void UpdateFields(Point position, (int R, int Q) coords)
        {
            _fields[coords.R, coords.Q].Update(position, HexagonRadius);
        }

        private void UpdateRoadSpots(Point pos, (int R, int Q) coords)
        {
            int maxId = _viewModel.BoardSize * _viewModel.BoardSize * 2;
            double marginOffset = EllipseDiameter / 2;
            int[] nodeIds = new int[]
            {
                GetNodeId(coords.R - 1, coords.Q + 1, 0),
                GetNodeId(coords.R + 1, coords.Q, 1),
                GetNodeId(coords.R, coords.Q, 0),
                GetNodeId(coords.R + 1, coords.Q - 1, 1),
            };
            for (int i = 0; i < 3; ++i)
            {
                Point firstNode = HexagonFactory.HexCorner(pos, HexagonRadius, i);
                Point secondNode = HexagonFactory.HexCorner(pos, HexagonRadius, i + 1);
                Point roadPos = new Point((firstNode.X + secondNode.X) / 2, (firstNode.Y + secondNode.Y) / 2);
                
                int firstNodeId = nodeIds[i];
                int secondNodeId = nodeIds[i + 1];

                if (firstNodeId >= 0 && secondNodeId >= 0 && firstNodeId < maxId && secondNodeId < maxId)
                    _roads[firstNodeId, secondNodeId].Margin = new Thickness(roadPos.X - marginOffset, roadPos.Y - marginOffset, 0, 0);
            }
        }

        private void UpdateBuildingSpots()
        { 
            double ellipseMarginOffset = EllipseDiameter / 2;
            int[] cornerIds = { 2, 5 };
            foreach (NodeViewModel node in _viewModel.Nodes)
            {
                if (node is null) continue;

                (int R, int Q, int D) coords = node.Position;
                Point position = CalculatePositionOnCanvas(coords.R, coords.Q);
                Point cornerPos = HexagonFactory.HexCorner(position, HexagonRadius, cornerIds[coords.D]);

                _settlements[coords.R, coords.Q, coords.D].Margin = new Thickness(cornerPos.X - ellipseMarginOffset, cornerPos.Y - ellipseMarginOffset, 0, 0);
            }
        }

        private void CreateFields(Point position, (int R, int Q) coords)
        {
            //Polygon field;
            //field = HexagonFactory.MakeHexagon(position, HexagonRadius);
            //field.Stroke = Brushes.Black;
            //field.Fill = GetBrush(_viewModel.Fields[coords.R * _viewModel.BoardSize + coords.Q].FieldType);
            //field.StrokeThickness = StrokeThickness;

            //Label label = new Label();
            //label.Content = coords.Q.ToString() + "," + coords.R.ToString();
            //label.Margin = new Thickness(position.X - HexagonRadius / 2, position.Y - HexagonRadius / 2, 0, 0);

            CatanField field = new CatanField(position, coords, HexagonRadius);

            field.Field.DataContext = _viewModel.Fields[coords.R * _viewModel.BoardSize + coords.Q];
            field.Field.Style = this.Resources["FieldStyle"] as Style;

            Children.Add(field.Field);
            Children.Add(field.Text);
            _fields[coords.R, coords.Q] = field;
        }

        private void CreateRoadSpots(Point pos, (int R, int Q) coords) 
        {
            int maxId = _viewModel.BoardSize * _viewModel.BoardSize * 2;
            double marginOffset = EllipseDiameter / 2;
            int[] nodeIds = new int[]
            {
                GetNodeId(coords.R - 1, coords.Q + 1, 0),
                GetNodeId(coords.R + 1, coords.Q, 1),
                GetNodeId(coords.R, coords.Q, 0),
                GetNodeId(coords.R + 1, coords.Q - 1, 1),
            };
            for (int i = 0; i < 3; ++i)
            {
                Point firstNode = HexagonFactory.HexCorner(pos, HexagonRadius, i);
                Point secondNode = HexagonFactory.HexCorner(pos, HexagonRadius, i + 1);

                Point roadPos = new Point((firstNode.X + secondNode.X) / 2, (firstNode.Y + secondNode.Y) / 2);
                //object res = CatanFieldTemplate.FindName("Ellipse", this);
                Ellipse roadEllipse = new Ellipse()
                {
                    Width = EllipseDiameter,
                    Height = EllipseDiameter,
                    Margin = new Thickness(roadPos.X - marginOffset, roadPos.Y - marginOffset, 0, 0),
                    Fill = Brushes.LightGray,
                    Stroke = Brushes.Green,
                    StrokeThickness = 1,
                    Opacity = 0.5,
                    Visibility = Visibility.Hidden
                };

                roadEllipse.MouseDown += Ellipse_Click;
                SetZIndex(roadEllipse, 100);
                Children.Add(roadEllipse);
                int firstNodeId = nodeIds[i];
                int secondNodeId = nodeIds[i + 1];

                if (firstNodeId >= 0 && secondNodeId >= 0 && firstNodeId < maxId && secondNodeId < maxId)
                {
                    _roads[firstNodeId, secondNodeId] = roadEllipse;
                    _roads[secondNodeId, firstNodeId] = roadEllipse;
                    _viewModel.Roads[firstNodeId][secondNodeId].PropertyChanged +=
                        (obj, param) => UpdateRoad(firstNodeId, secondNodeId);
                }
            }
        }

        private void CreateBuildingSpots()
        {
            double ellipseMarginOffset = EllipseDiameter / 2;
            int[] cornerIds = { 2, 5 };
            foreach (NodeViewModel node in _viewModel.Nodes)
            {
                if (node is null) continue;

                (int R, int Q, int D) coords =  node.Position;
                Point position = CalculatePositionOnCanvas(coords.R, coords.Q);
                Point cornerPos = HexagonFactory.HexCorner(position, HexagonRadius, cornerIds[coords.D]);
                Ellipse ellipse = new Ellipse();
                ellipse.Margin = new Thickness(cornerPos.X - ellipseMarginOffset, cornerPos.Y - ellipseMarginOffset, 0, 0);

                SetZIndex(ellipse, 100);
                int coordinateId = GetNodeId(coords.R, coords.Q, coords.D);
                ellipse.DataContext = _viewModel.Nodes[coordinateId];
                ellipse.Style = this.Resources["NodeStyle"] as Style;
                ellipse.MouseDown += (obj, args) => node.UpgradeNodeCommand.Execute(null);
                _settlements[coords.R, coords.Q, coords.D] = ellipse;
                Children.Add(ellipse);
            }           
        }

        private Point CalculatePositionOnCanvas(int r, int q)
        {
            int offsetR = r - Offset;
            int offsetQ = q - Offset;
            double xCoord = WindowSize / 2 + HexagonRadius * (Math.Sqrt(3) * offsetQ + Math.Sqrt(3) * offsetR * 0.5);
            double yCoord = WindowSize / 2 + HexagonRadius * 1.5 * offsetR;
            return  new Point(xCoord, yCoord);
        }

        private Brush GetBuildingBrush(String buildingType)
        {
            switch(buildingType)
            {
                case "Settlement":
                    return Brushes.Red;
                case "City":
                    return Brushes.Gold;
                default:
                    return Brushes.LightGray;
            }
        }

        private void UpdateRoad(int firstNodeId, int secondNodeId)
        {
            _roads[firstNodeId, secondNodeId].Visibility = _viewModel.Roads[firstNodeId][secondNodeId].IsBuildable ? Visibility.Visible : Visibility.Hidden;
        }
        private void Ellipse_Click(object obj, MouseButtonEventArgs e)
        {
            Ellipse house = obj as Ellipse;
            Thickness t = house.Margin;
            MessageBox.Show(((int)t.Left).ToString() + ", " + ((int)t.Top).ToString());
        }

        private int GetNodeId(int r, int q, int dir)
        {
            return r * _viewModel.BoardSize * 2 + q * 2 + dir;
        }

        private int GetNodeId((int R, int Q, int Dir)coords)
        {
            return GetNodeId(coords.R, coords.Q, coords.Dir);
        }
        #endregion
    }
}
