﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using CatanGame.Model;

namespace CatanGame.ViewModel
{
    public class CatanViewModel : ViewModelBase
    {
        #region Private fields
        private readonly GameModel _model;
        #endregion

        #region Public properties
        public ObservableCollection<FieldViewModel> Fields { get; set; }
        public ObservableCollection<ObservableCollection<RoadViewModel>> Roads { get; set; }
        public ObservableCollection<NodeViewModel> Nodes { get; set; }

        public int BoardSize { get => _model.Board.Size; }
        public int NodeCount { get => _model.Board.VertexCount; }
        #endregion

        #region Public constructors

        public CatanViewModel(GameModel model)
        {
            //model csatlakozatatasa
            _model = model;
            _model.NodeUpgraded += Model_NodeChanged;
            _model.GameRolled += new EventHandler(Model_Rolled);

            //parancsok kezelese

            Init();
        }

        #endregion

        #region Private methods
        private void Init()
        {
            // initialize commands
            ShowEnabledRoadsCommand = new DelegateCommand(ShowEnabledRoads);

            // initialize collections
            InitFields();
            InitNodes();
            InitRoads();
        }

        private void InitFields()
        {
            Fields = new();
            FieldViewModel field = null;
            for (int r = 0; r < _model.Board.Size; ++r)
            {
                for (int q = 0; q < _model.Board.Size; ++q)
                {
                    field = null;
                    if (_model.Board.Fields[r, q] is not null)
                    {
                        field = new FieldViewModel()
                        {
                            R = r,
                            Q = q,
                            FieldNumber = 0,
                            FieldType = _model.Board.Fields[r, q].ResourceType.ToString(),
                            HasRobber = false,
                        };

                    }
                    Fields.Add(field);
                }
            }
        }

        private void InitNodes()
        {
            Nodes = new();
            for (int r = 0; r < _model.Board.Size; ++r)
            {
                for (int q = 0; q < _model.Board.Size; ++q)
                {
                    for (int dir = 0; dir < 2; ++dir)
                    {
                        NodeViewModel node = null;

                        Node boardNode = _model.Board.Nodes[r, q, dir];
                        if (boardNode is not null)
                        {
                            NodeCoordinates coords = new NodeCoordinates(r, q, (Direction) dir);
                            node = new NodeViewModel()
                            {
                                //HouseOwner = boardNode.Building.Owner.Name,
                                HouseType = boardNode.Building.Type.ToString(),
                                Position = (r, q, dir),
                                UpgradeNodeCommand = new DelegateCommand((param) => UpgradeNode?.Invoke(this, coords))
                            };

                        }
                        Nodes.Add(node);
                    }
                }
            }
        }
  
        private void InitRoads()
        {
            int totalNodeCount = _model.Board.Size * _model.Board.Size * 2;
            Roads = new();
            // Note: iterating vertically to make symmetric matrix share references of Roads
            for (int i = 0; i < totalNodeCount; ++i)
                Roads.Add(new());

            for (int j = 0; j < totalNodeCount; ++j)
            {
                for (int i = 0; i < totalNodeCount; ++i)
                {
                    if (j > i)
                        Roads[i].Add(Roads[j][i]);
                    else
                        Roads[i].Add(new RoadViewModel()
                        {
                            IsBuildable = false,
                            Position = (i, j)
                        });
                }
            }
        }

        private void ShowEnabledRoads(object param)
        {
            List<(NodeCoordinates, NodeCoordinates)> enabledRoads = _model.EnabledRoads(_model.CurrentPlayerId);

            // setting IsBuildable to true at the reverse coordinates is not needed, because 
            // the normal coordinates and the reverse coordinates share the same reference.
           foreach ((NodeCoordinates FirstNode, NodeCoordinates SecondNode) road in enabledRoads)
                Roads[GetNodeId(road.FirstNode)][GetNodeId(road.SecondNode)].IsBuildable = true;
        }
        private int GetNodeId(int r, int q, int dir)
        {
            return r * BoardSize * 2 + q * 2 + dir;
        }

        private int GetNodeId((int R, int Q, int Dir) coords)
        {
            return GetNodeId(coords.R, coords.Q, coords.Dir);
        }

        private int GetNodeId(NodeCoordinates coords)
        {
            return GetNodeId(coords.R, coords.Q, (int) coords.Dir);
        }

        #endregion

        #region Public commands
        public DelegateCommand NewGameCommand { get; private set; }
        public DelegateCommand ExitCommand { get; private set; }
        public DelegateCommand BuyCardCommand { get; private set; }
        public DelegateCommand PlaceRoadCommand { get; private set; }
        public DelegateCommand PlaceSettlementCommand { get; private set; }
        public DelegateCommand UpgradeToCityCommand { get; private set; }
        public DelegateCommand StepGameCommand { get; private set; }
        public DelegateCommand RollCommand { get; private set; }
        public DelegateCommand ShowEnabledRoadsCommand { get; private set; }
        #endregion

        #region Public methods
        public List<(int, int)> EnableRoads()
        {
            List<(NodeCoordinates FirstCoords, NodeCoordinates SecondCoords)> roads = _model.EnabledRoads(_model.CurrentPlayerId);
            return roads.Select(coords => {
                return (GetNodeId(coords.FirstCoords), GetNodeId(coords.SecondCoords));
            }).ToList();
        }
        #endregion




        #region Model event handlers
        private void Model_CardBought(object obj, EventArgs e)
        {

        }
        private void Model_RoadPlaced(object obj, EventArgs e)
        {

        }
        private void Model_NodeChanged(object obj, NodeCoordinates coords)
        {
            int nodeId = coords.R * BoardSize * 2 + coords.Q * 2 + (int)coords.Dir;
            Node node = _model.Board.Nodes[coords.R, coords.Q, (int)coords.Dir];

            Nodes[nodeId].HouseType = node.Building.Type.ToString();
        }
        private void Model_GameStepped(object obj, EventArgs e)
        {

        }
        private void Model_GameOver(object obj, EventArgs e)
        {

        }
        private void Model_Rolled(object obj, EventArgs e)
        {

        }

        private void Model_ShowEnabledRoads(object ibj, EventArgs e)
        {
            
        }


        #endregion

        #region Public events

        public event EventHandler<EventArgs> NewGame;
        public event EventHandler<EventArgs> Exit;
        public event EventHandler<EventArgs> BuyCard;
        public event EventHandler<EventArgs> PlaceRoad;
        public event EventHandler<NodeCoordinates> UpgradeNode;
        public event EventHandler<EventArgs> UpgradeToCity;
        public event EventHandler<EventArgs> StepGame;
        public event EventHandler<EventArgs> Roll;

        #endregion
    }
}
